# ------------------------------------------------------------------------------
#  es7s/core
#  (c) 2024 A. Shavykin <0.delameter@gmail.com>
# ------------------------------------------------------------------------------
from dataclasses import dataclass

import pytermor as pt
from pytermor import ExtendedEnum

from ..shared import Styles as BaseStyles


@dataclass
class EncodingValues:
    label: str
    style: pt.FT = None

    def __hash__(self):
        return hash(self.label) + hash(self.style)


class Encoding(EncodingValues, ExtendedEnum):
    KOI8_R = EncodingValues("koi8-r", pt.cv.RED)
    CP_437 = EncodingValues("cp437", pt.cv.HI_YELLOW)
    CP_866 = EncodingValues("cp866", pt.cv.BLUE)
    ISO_8859_5 = EncodingValues("iso8859-5", pt.cv.GREEN)
    WIN_1251 = EncodingValues("windows-1251", pt.cv.GRAY_50)
    WIN_1252 = EncodingValues("windows-1252", pt.cv.CYAN)
    MAC = EncodingValues("macintosh", pt.cv.MAGENTA)
    UTF_8 = EncodingValues("utf-8", pt.cv.ORANGE_1)

    @classmethod
    def is_allowed(cls, cfrom: "Encoding" = None, cto: "Encoding" = None) -> bool:
        return cfrom != cto and cto != Encoding.UTF_8 and cfrom != Encoding.WIN_1252

    def format(self, width: int = 15, align: str = None) -> pt.Fragment:
        return pt.Fragment(pt.fit(self.value.label, width, align=align), BaseStyles.TEXT_DEFAULT)

    def format_marker(self, char="•") -> pt.Fragment:
        return pt.Fragment(char, self.value.style)


class EncodingPair(tuple[Encoding, Encoding]):
    def format(self, reverse=False) -> pt.Text:
        if reverse:
            return pt.Text(self[0].format(align=">"), " ← ", self[1].format())
        return pt.Text(self[0].format(align=">"), " → ", self[1].format())

    def generate_code(self) -> str:
        record = "("
        record += ", ".join(f"Encoding.{enc.name}" for enc in self)
        record += ")"
        return record


@dataclass(frozen=True)
class EncodingStats:
    encpair: EncodingPair
    char: str
    freq: float


# --------------------------------------------------
# CODE GENERATED
ENCODINGS_MOST_FREQ_CHARS = {
    (Encoding.KOI8_R, Encoding.CP_437): [
        ("\u2567", 10.547),
        ("\u253c", 7.680),
        ("\u2534", 7.518),
        ("\u2554", 6.128),
        ("\u256c", 5.900),
        ("\u2558", 5.367),
        ("\u2559", 4.834),
        ("\u2560", 4.727),
        ("\u2565", 4.096),
        ("\u256b", 4.089),
        ("\u2566", 3.105),
        ("\u002c", 2.780),
        ("\u2500", 2.739),
        ("\u2550", 2.690),
        ("\u2552", 2.585),
        ("\u2568", 2.234),
    ],
    (Encoding.KOI8_R, Encoding.CP_866): [
        ("\u2567", 10.547),
        ("\u253c", 7.680),
        ("\u2534", 7.518),
        ("\u2554", 6.128),
        ("\u256c", 5.900),
        ("\u2558", 5.367),
        ("\u2559", 4.834),
        ("\u2560", 4.727),
        ("\u2565", 4.096),
        ("\u256b", 4.089),
        ("\u2566", 3.105),
        ("\u002c", 2.780),
        ("\u2500", 2.739),
        ("\u2550", 2.690),
        ("\u2552", 2.585),
        ("\u2568", 2.234),
    ],
    (Encoding.KOI8_R, Encoding.ISO_8859_5): [
        ("\u042f", 10.547),
        ("\u0425", 7.680),
        ("\u0421", 7.518),
        ("\u0429", 6.128),
        ("\u042e", 5.900),
        ("\u0434", 5.367),
        ("\u0433", 4.834),
        ("\u042c", 4.727),
        ("\u0432", 4.096),
        ("\u0437", 4.089),
        ("\u042b", 3.105),
        ("\u002c", 2.780),
        ("\u0424", 2.739),
        ("\u042d", 2.690),
        ("\u0435", 2.585),
        ("\u0430", 2.234),
    ],
    (Encoding.KOI8_R, Encoding.WIN_1251): [
        ("\u041f", 10.547),
        ("\u0415", 7.680),
        ("\u0411", 7.518),
        ("\u0419", 6.128),
        ("\u041e", 5.900),
        ("\u0424", 5.367),
        ("\u0423", 4.834),
        ("\u041c", 4.727),
        ("\u0422", 4.096),
        ("\u0427", 4.089),
        ("\u041b", 3.105),
        ("\u002c", 2.780),
        ("\u0414", 2.739),
        ("\u041d", 2.690),
        ("\u0425", 2.585),
        ("\u0420", 2.234),
    ],
    (Encoding.KOI8_R, Encoding.WIN_1252): [
        ("\u00cf", 10.547),
        ("\u00c5", 7.680),
        ("\u00c1", 7.518),
        ("\u00c9", 6.128),
        ("\u00ce", 5.900),
        ("\u00d4", 5.367),
        ("\u00d3", 4.834),
        ("\u00cc", 4.727),
        ("\u00d2", 4.096),
        ("\u00d7", 4.089),
        ("\u00cb", 3.105),
        ("\u002c", 2.780),
        ("\u00c4", 2.739),
        ("\u00cd", 2.690),
        ("\u00d5", 2.585),
        ("\u00d0", 2.234),
    ],
    (Encoding.KOI8_R, Encoding.MAC): [
        ("\u0153", 10.547),
        ("\u2248", 7.680),
        ("\u00a1", 7.518),
        ("\u2026", 6.128),
        ("\u0152", 5.900),
        ("\u2018", 5.367),
        ("\u201d", 4.834),
        ("\u00c3", 4.727),
        ("\u201c", 4.096),
        ("\u25ca", 4.089),
        ("\u00c0", 3.105),
        ("\u002c", 2.780),
        ("\u0192", 2.739),
        ("\u00d5", 2.690),
        ("\u2019", 2.585),
        ("\u2013", 2.234),
    ],
    (Encoding.CP_866, Encoding.KOI8_R): [
        ("\u255d", 10.547),
        ("\u2554", 7.680),
        ("\u2550", 7.518),
        ("\u2557", 6.128),
        ("\u255c", 5.900),
        ("\u0411", 5.367),
        ("\u0410", 4.834),
        ("\u255a", 4.727),
        ("\u042e", 4.096),
        ("\u2552", 4.089),
        ("\u2559", 3.105),
        ("\u002c", 2.780),
        ("\u2553", 2.739),
        ("\u255b", 2.690),
        ("\u0426", 2.585),
        ("\u255e", 2.234),
    ],
    (Encoding.CP_866, Encoding.CP_437): [
        ("\u00ab", 10.547),
        ("\u00d1", 7.680),
        ("\u00e1", 7.518),
        ("\u00bf", 6.128),
        ("\u00a1", 5.900),
        ("\u0393", 5.367),
        ("\u00df", 4.834),
        ("\u00bd", 4.727),
        ("\u03b1", 4.096),
        ("\u00f3", 4.089),
        ("\u00ac", 3.105),
        ("\u002c", 2.780),
        ("\u00f1", 2.739),
        ("\u00bc", 2.690),
        ("\u03c0", 2.585),
        ("\u00bb", 2.234),
    ],
    (Encoding.CP_866, Encoding.ISO_8859_5): [
        ("\u040e", 10.547),
        ("\u0405", 7.680),
        ("\u00a0", 7.518),
        ("\u0408", 6.128),
        ("\u00ad", 5.900),
        ("\u0442", 5.367),
        ("\u0441", 4.834),
        ("\u040b", 4.727),
        ("\u0440", 4.096),
        ("\u0402", 4.089),
        ("\u040a", 3.105),
        ("\u002c", 2.780),
        ("\u0404", 2.739),
        ("\u040c", 2.690),
        ("\u0443", 2.585),
        ("\u040f", 2.234),
    ],
    (Encoding.CP_866, Encoding.WIN_1251): [
        ("\u00ae", 10.547),
        ("\u0490", 7.680),
        ("\u00a0", 7.518),
        ("\u0401", 6.128),
        ("\u00ad", 5.900),
        ("\u0432", 5.367),
        ("\u0431", 4.834),
        ("\u00ab", 4.727),
        ("\u0430", 4.096),
        ("\u045e", 4.089),
        ("\u0404", 3.105),
        ("\u002c", 2.780),
        ("\u00a4", 2.739),
        ("\u00ac", 2.690),
        ("\u0433", 2.585),
        ("\u0407", 2.234),
    ],
    (Encoding.CP_866, Encoding.WIN_1252): [
        ("\u00ae", 10.547),
        ("\u00a5", 7.680),
        ("\u00a0", 7.518),
        ("\u00a8", 6.128),
        ("\u00ad", 5.900),
        ("\u00e2", 5.367),
        ("\u00e1", 4.834),
        ("\u00ab", 4.727),
        ("\u00e0", 4.096),
        ("\u00a2", 4.089),
        ("\u00aa", 3.105),
        ("\u002c", 2.780),
        ("\u00a4", 2.739),
        ("\u00ac", 2.690),
        ("\u00e3", 2.585),
        ("\u00af", 2.234),
    ],
    (Encoding.CP_866, Encoding.MAC): [
        ("\u00c6", 10.547),
        ("\u2022", 7.680),
        ("\u2020", 7.518),
        ("\u00ae", 6.128),
        ("\u2260", 5.900),
        ("\u201a", 5.367),
        ("\u00b7", 4.834),
        ("\u00b4", 4.727),
        ("\u2021", 4.096),
        ("\u00a2", 4.089),
        ("\u2122", 3.105),
        ("\u002c", 2.780),
        ("\u00a7", 2.739),
        ("\u00a8", 2.690),
        ("\u201e", 2.585),
        ("\u00d8", 2.234),
    ],
    (Encoding.ISO_8859_5, Encoding.KOI8_R): [
        ("\u0447", 10.547),
        ("\u0443", 7.680),
        ("\u043f", 7.518),
        ("\u044c", 6.128),
        ("\u0449", 5.900),
        ("\u0411", 5.367),
        ("\u0410", 4.834),
        ("\u0448", 4.727),
        ("\u042e", 4.096),
        ("\u0440", 4.089),
        ("\u0437", 3.105),
        ("\u002c", 2.780),
        ("\u0442", 2.739),
        ("\u044d", 2.690),
        ("\u0426", 2.585),
        ("\u044a", 2.234),
    ],
    (Encoding.ISO_8859_5, Encoding.CP_437): [
        ("\u2590", 10.547),
        ("\u2552", 7.680),
        ("\u2568", 7.518),
        ("\u256a", 6.128),
        ("\u258c", 5.900),
        ("\u0393", 5.367),
        ("\u00df", 4.834),
        ("\u2588", 4.727),
        ("\u03b1", 4.096),
        ("\u2565", 4.089),
        ("\u250c", 3.105),
        ("\u002c", 2.780),
        ("\u2558", 2.739),
        ("\u2584", 2.690),
        ("\u03c0", 2.585),
        ("\u2580", 2.234),
    ],
    (Encoding.ISO_8859_5, Encoding.CP_866): [
        ("\u2590", 10.547),
        ("\u2552", 7.680),
        ("\u2568", 7.518),
        ("\u256a", 6.128),
        ("\u258c", 5.900),
        ("\u0442", 5.367),
        ("\u0441", 4.834),
        ("\u2588", 4.727),
        ("\u0440", 4.096),
        ("\u2565", 4.089),
        ("\u250c", 3.105),
        ("\u002c", 2.780),
        ("\u2558", 2.739),
        ("\u2584", 2.690),
        ("\u0443", 2.585),
        ("\u2580", 2.234),
    ],
    (Encoding.ISO_8859_5, Encoding.WIN_1251): [
        ("\u042e", 10.547),
        ("\u0425", 7.680),
        ("\u0420", 7.518),
        ("\u0428", 6.128),
        ("\u042d", 5.900),
        ("\u0432", 5.367),
        ("\u0431", 4.834),
        ("\u042b", 4.727),
        ("\u0430", 4.096),
        ("\u0422", 4.089),
        ("\u042a", 3.105),
        ("\u002c", 2.780),
        ("\u0424", 2.739),
        ("\u042c", 2.690),
        ("\u0433", 2.585),
        ("\u042f", 2.234),
    ],
    (Encoding.ISO_8859_5, Encoding.WIN_1252): [
        ("\u00de", 10.547),
        ("\u00d5", 7.680),
        ("\u00d0", 7.518),
        ("\u00d8", 6.128),
        ("\u00dd", 5.900),
        ("\u00e2", 5.367),
        ("\u00e1", 4.834),
        ("\u00db", 4.727),
        ("\u00e0", 4.096),
        ("\u00d2", 4.089),
        ("\u00da", 3.105),
        ("\u002c", 2.780),
        ("\u00d4", 2.739),
        ("\u00dc", 2.690),
        ("\u00e3", 2.585),
        ("\u00df", 2.234),
    ],
    (Encoding.ISO_8859_5, Encoding.MAC): [
        ("\ufb01", 10.547),
        ("\u2019", 7.680),
        ("\u2013", 7.518),
        ("\u00ff", 6.128),
        ("\u203a", 5.900),
        ("\u201a", 5.367),
        ("\u00b7", 4.834),
        ("\u20ac", 4.727),
        ("\u2021", 4.096),
        ("\u201c", 4.089),
        ("\u2044", 3.105),
        ("\u002c", 2.780),
        ("\u2018", 2.739),
        ("\u2039", 2.690),
        ("\u201e", 2.585),
        ("\ufb02", 2.234),
    ],
    (Encoding.WIN_1251, Encoding.KOI8_R): [
        ("\u041d", 10.547),
        ("\u0415", 7.680),
        ("\u042e", 7.518),
        ("\u0425", 6.128),
        ("\u041c", 5.900),
        ("\u0420", 5.367),
        ("\u042f", 4.834),
        ("\u041a", 4.727),
        ("\u041f", 4.096),
        ("\u0411", 4.089),
        ("\u0419", 3.105),
        ("\u002c", 2.780),
        ("\u0414", 2.739),
        ("\u041b", 2.690),
        ("\u0421", 2.585),
        ("\u041e", 2.234),
    ],
    (Encoding.WIN_1251, Encoding.CP_437): [
        ("\u03b5", 10.547),
        ("\u03c3", 7.680),
        ("\u03b1", 7.518),
        ("\u03a6", 6.128),
        ("\u03c6", 5.900),
        ("\u2265", 5.367),
        ("\u00b1", 4.834),
        ("\u03b4", 4.727),
        ("\u2261", 4.096),
        ("\u0393", 4.089),
        ("\u03a9", 3.105),
        ("\u002c", 2.780),
        ("\u03a3", 2.739),
        ("\u221e", 2.690),
        ("\u2264", 2.585),
        ("\u2229", 2.234),
    ],
    (Encoding.WIN_1251, Encoding.CP_866): [
        ("\u044e", 10.547),
        ("\u0445", 7.680),
        ("\u0440", 7.518),
        ("\u0448", 6.128),
        ("\u044d", 5.900),
        ("\u0404", 5.367),
        ("\u0451", 4.834),
        ("\u044b", 4.727),
        ("\u0401", 4.096),
        ("\u0442", 4.089),
        ("\u044a", 3.105),
        ("\u002c", 2.780),
        ("\u0444", 2.739),
        ("\u044c", 2.690),
        ("\u0454", 2.585),
        ("\u044f", 2.234),
    ],
    (Encoding.WIN_1251, Encoding.ISO_8859_5): [
        ("\u044e", 10.547),
        ("\u0445", 7.680),
        ("\u0440", 7.518),
        ("\u0448", 6.128),
        ("\u044d", 5.900),
        ("\u0452", 5.367),
        ("\u0451", 4.834),
        ("\u044b", 4.727),
        ("\u2116", 4.096),
        ("\u0442", 4.089),
        ("\u044a", 3.105),
        ("\u002c", 2.780),
        ("\u0444", 2.739),
        ("\u044c", 2.690),
        ("\u0453", 2.585),
        ("\u044f", 2.234),
    ],
    (Encoding.WIN_1251, Encoding.WIN_1252): [
        ("\u00ee", 10.547),
        ("\u00e5", 7.680),
        ("\u00e0", 7.518),
        ("\u00e8", 6.128),
        ("\u00ed", 5.900),
        ("\u00f2", 5.367),
        ("\u00f1", 4.834),
        ("\u00eb", 4.727),
        ("\u00f0", 4.096),
        ("\u00e2", 4.089),
        ("\u00ea", 3.105),
        ("\u002c", 2.780),
        ("\u00e4", 2.739),
        ("\u00ec", 2.690),
        ("\u00f3", 2.585),
        ("\u00ef", 2.234),
    ],
    (Encoding.WIN_1251, Encoding.MAC): [
        ("\u00d3", 10.547),
        ("\u00c2", 7.680),
        ("\u2021", 7.518),
        ("\u00cb", 6.128),
        ("\u00cc", 5.900),
        ("\u00da", 5.367),
        ("\u00d2", 4.834),
        ("\u00ce", 4.727),
        ("\uf8ff", 4.096),
        ("\u201a", 4.089),
        ("\u00cd", 3.105),
        ("\u002c", 2.780),
        ("\u2030", 2.739),
        ("\u00cf", 2.690),
        ("\u00db", 2.585),
        ("\u00d4", 2.234),
    ],
    (Encoding.UTF_8, Encoding.KOI8_R): [
        ("\u043f", 34.276),
        ("\u044f", 14.074),
        ("\u256c", 5.447),
        ("\u2563", 3.966),
        ("\u255f", 3.883),
        ("\u2566", 3.165),
        ("\u256b", 3.047),
        ("\u250c", 2.772),
        ("\u2502", 2.497),
        ("\u2569", 2.442),
        ("\u2500", 2.115),
        ("\u2561", 2.112),
        ("\u2568", 1.604),
        ("\u002c", 1.436),
        ("\u2562", 1.415),
        ("\u256a", 1.390),
    ],
    (Encoding.UTF_8, Encoding.CP_437): [
        ("\u2568", 34.276),
        ("\u2564", 14.074),
        ("\u255b", 5.447),
        ("\u2561", 3.966),
        ("\u2591", 3.883),
        ("\u2555", 3.165),
        ("\u255c", 3.047),
        ("\u00e9", 2.772),
        ("\u00fc", 2.497),
        ("\u2557", 2.442),
        ("\u00c7", 2.115),
        ("\u2593", 2.112),
        ("\u2551", 1.604),
        ("\u002c", 1.436),
        ("\u2524", 1.415),
        ("\u255d", 1.390),
    ],
    (Encoding.UTF_8, Encoding.CP_866): [
        ("\u2568", 34.276),
        ("\u2564", 14.074),
        ("\u255b", 5.447),
        ("\u2561", 3.966),
        ("\u2591", 3.883),
        ("\u2555", 3.165),
        ("\u255c", 3.047),
        ("\u0412", 2.772),
        ("\u0411", 2.497),
        ("\u2557", 2.442),
        ("\u0410", 2.115),
        ("\u2593", 2.112),
        ("\u2551", 1.604),
        ("\u002c", 1.436),
        ("\u2524", 1.415),
        ("\u255d", 1.390),
    ],
    (Encoding.UTF_8, Encoding.ISO_8859_5): [
        ("\u0430", 34.276),
        ("\u0431", 14.074),
        ("\u041e", 5.447),
        ("\u0415", 3.966),
        ("\u0410", 3.883),
        ("\u0418", 3.165),
        ("\u041d", 3.047),
        ("\u0082", 2.772),
        ("\u0081", 2.497),
        ("\u041b", 2.442),
        ("\u0080", 2.115),
        ("\u0412", 2.112),
        ("\u041a", 1.604),
        ("\u002c", 1.436),
        ("\u0414", 1.415),
        ("\u041c", 1.390),
    ],
    (Encoding.UTF_8, Encoding.WIN_1251): [
        ("\u0420", 34.276),
        ("\u0421", 14.074),
        ("\u0455", 5.447),
        ("\u00b5", 3.966),
        ("\u00b0", 3.883),
        ("\u0451", 3.165),
        ("\u0405", 3.047),
        ("\u201a", 2.772),
        ("\u0403", 2.497),
        ("\u00bb", 2.442),
        ("\u0402", 2.115),
        ("\u0406", 2.112),
        ("\u0454", 1.604),
        ("\u002c", 1.436),
        ("\u0491", 1.415),
        ("\u0458", 1.390),
    ],
    (Encoding.UTF_8, Encoding.WIN_1252): [
        ("\u00d0", 34.276),
        ("\u00d1", 14.074),
        ("\u00be", 5.447),
        ("\u00b5", 3.966),
        ("\ufffd", 3.918),
        ("\u00b0", 3.883),
        ("\u00b8", 3.165),
        ("\u00bd", 3.047),
        ("\u201a", 2.772),
        ("\u00bb", 2.442),
        ("\u20ac", 2.115),
        ("\u00b2", 2.112),
        ("\u00ba", 1.604),
        ("\u002c", 1.436),
        ("\u00b4", 1.415),
        ("\u00bc", 1.390),
    ],
    (Encoding.UTF_8, Encoding.MAC): [
        ("\u2013", 34.276),
        ("\u2014", 14.074),
        ("\u00e6", 5.447),
        ("\u00b5", 3.966),
        ("\u221e", 3.883),
        ("\u220f", 3.165),
        ("\u03a9", 3.047),
        ("\u00c7", 2.772),
        ("\u00c5", 2.497),
        ("\u00aa", 2.442),
        ("\u00c4", 2.115),
        ("\u2264", 2.112),
        ("\u222b", 1.604),
        ("\u002c", 1.436),
        ("\u00a5", 1.415),
        ("\u00ba", 1.390),
    ],
}
MOST_FREQ_CHAR_MAP = {
    "\u2567": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u253c": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u2534": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u2554": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u256c": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2558": [
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u2559": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u2560": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u2565": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u256b": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2566": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u002c": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u2500": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2550": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u2552": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u2568": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2564": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u256a": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u2518": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u255f": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u250c": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u252c": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u002d": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u2590": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u002e": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u2569": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2553": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u2588": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u255a": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u2514": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
    ],
    "\u000a": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u251c": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u258c": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u2584": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u03b5": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u2261": [(Encoding.KOI8_R, Encoding.CP_437), (Encoding.WIN_1251, Encoding.CP_437)],
    "\u00df": [
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
    ],
    "\u255e": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u2229": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u2248": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u03b4": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u0393": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
    ],
    "\u003f": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0022": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u03c6": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u03a3": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u2264": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0398": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u2320": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u2265": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u0021": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u003a": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00b1": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u207f": [(Encoding.KOI8_R, Encoding.CP_437), (Encoding.WIN_1251, Encoding.CP_437)],
    "\u2580": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u03c4": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u25a0": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
    ],
    "\u03c3": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u003b": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0029": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0028": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u221e": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00b7": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u2321": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u0061": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0049": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0031": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00f7": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00b5": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0030": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u03a6": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u221a": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
    ],
    "\u0058": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0065": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0038": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0056": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0032": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0035": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0036": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0079": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0033": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u03c0": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
    ],
    "\u006f": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0072": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0073": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0037": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0075": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0034": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0039": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u006d": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0074": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u006c": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u006e": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0069": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00fa": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
    ],
    "\u0064": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0062": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u004e": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0063": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u03b1": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_437),
    ],
    "\u0027": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u005c": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0042": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0070": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0076": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u004c": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0078": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u003c": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0041": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00b0": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u0068": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0046": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0050": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0048": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0053": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0066": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u004d": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0043": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u004a": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0045": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0044": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u002a": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u004f": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0055": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u003e": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00b2": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u03a9": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u005e": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0067": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u004b": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0054": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0077": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u003d": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0047": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u0052": [
        (Encoding.KOI8_R, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u044e": [
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0401": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.WIN_1251),
    ],
    "\u0441": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u044f": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u045e": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.WIN_1251),
    ],
    "\u044b": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0442": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
    ],
    "\u044d": [
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0444": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0454": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0449": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
    ],
    "\u0407": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
    ],
    "\u0404": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.CP_866),
    ],
    "\u0451": [
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u2116": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0447": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
    ],
    "\u0445": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u044c": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
    ],
    "\u0457": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u040e": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u0446": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0448": [
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0443": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
    ],
    "\u0433": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0440": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u00a4": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u044a": [
        (Encoding.KOI8_R, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u042f": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0425": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0421": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0429": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u042e": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0434": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u042c": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0432": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0437": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u042b": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0424": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u042d": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0435": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0430": [
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0431": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0438": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0439": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0427": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u043a": [
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u0422": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u043e": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u042a": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0436": [
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
    ],
    "\u043b": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0428": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0420": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0423": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u043d": [
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
    ],
    "\u043c": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
    ],
    "\u0426": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
    ],
    "\u0453": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0452": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u045c": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u043f": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u045a": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0455": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0456": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u045b": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u0403": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u0458": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u00a7": [
        (Encoding.KOI8_R, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u041f": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u0415": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0411": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u0419": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u041e": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u041c": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
    ],
    "\u041b": [
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0414": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u041d": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
    ],
    "\u0417": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.KOI8_R, Encoding.WIN_1251),
    ],
    "\u0412": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u041a": [
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.WIN_1251, Encoding.KOI8_R),
    ],
    "\u0418": [
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0410": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0413": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u0416": [
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
    ],
    "\u0408": [
        (Encoding.KOI8_R, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u00cf": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00c5": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00c1": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00c9": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00ce": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00d4": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00d3": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00cc": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00d2": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00d7": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00cb": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00c4": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00cd": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00d5": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00d0": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u00d1": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u00d8": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00d9": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00c7": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00da": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00c2": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00de": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00ca": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
    ],
    "\u00d6": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00db": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00c8": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00c0": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u00c3": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u00dd": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00dc": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00ee": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f0": [(Encoding.KOI8_R, Encoding.WIN_1252), (Encoding.WIN_1251, Encoding.WIN_1252)],
    "\u00e1": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
    ],
    "\u00c6": [
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u00ef": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00eb": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00e2": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00ed": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00e4": [
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f3": [
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
    ],
    "\u00e9": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
    ],
    "\u00f4": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f2": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f1": [
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00fc": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
    ],
    "\u00e7": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00fe": [(Encoding.KOI8_R, Encoding.WIN_1252), (Encoding.WIN_1251, Encoding.WIN_1252)],
    "\u00e5": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00ec": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f5": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f6": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00e6": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00e8": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00fb": [
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00e3": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
    ],
    "\u00a3": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u00e0": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u00f8": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00fd": [(Encoding.KOI8_R, Encoding.WIN_1252), (Encoding.WIN_1251, Encoding.WIN_1252)],
    "\u00ea": [
        (Encoding.KOI8_R, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u0153": [
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u00a1": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u2026": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u0152": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u2018": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u201d": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u201c": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u25ca": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u0192": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u2019": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u2013": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u2014": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00ff": [
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u0178": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u00ab": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
    ],
    "\u2044": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.ISO_8859_5, Encoding.MAC)],
    "\u00ac": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
    ],
    "\ufb01": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u00a0": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u20ac": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u00bb": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u00bf": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
    ],
    "\u203a": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u2039": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\uf8ff": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.WIN_1251, Encoding.MAC)],
    "\u2206": [
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.KOI8_R, Encoding.MAC),
    ],
    "\u02dc": [
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u201a": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u2030": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00b8": [
        (Encoding.WIN_1251, Encoding.WIN_1252),
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\ufb02": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u02db": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.WIN_1251, Encoding.MAC)],
    "\u02d9": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.WIN_1251, Encoding.MAC)],
    "\u0131": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.WIN_1251, Encoding.MAC)],
    "\u02c6": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u02da": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.WIN_1251, Encoding.MAC)],
    "\u201e": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
    ],
    "\u2021": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.MAC),
    ],
    "\u00af": [
        (Encoding.KOI8_R, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u02dd": [(Encoding.KOI8_R, Encoding.MAC), (Encoding.WIN_1251, Encoding.MAC)],
    "\u255d": [
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u2557": [
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
    ],
    "\u255c": [
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u255b": [
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2556": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2551": [
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2555": [
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2502": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2592": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2593": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2591": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2510": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2524": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2219": [
        (Encoding.CP_866, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.WIN_1251, Encoding.CP_437),
        (Encoding.WIN_1251, Encoding.CP_866),
    ],
    "\u00bd": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u00bc": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u00ba": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u2310": [(Encoding.UTF_8, Encoding.CP_437), (Encoding.CP_866, Encoding.CP_437)],
    "\u00aa": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u00a5": [
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u00f9": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.CP_866, Encoding.CP_437),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.WIN_1251, Encoding.WIN_1252),
    ],
    "\u20a7": [(Encoding.CP_866, Encoding.CP_437), (Encoding.UTF_8, Encoding.CP_437)],
    "\u0405": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u00ad": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u040b": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u0402": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u040a": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u040c": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u040f": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u0409": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
    ],
    "\u0406": [
        (Encoding.UTF_8, Encoding.ISO_8859_5),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.ISO_8859_5),
        (Encoding.UTF_8, Encoding.WIN_1251),
    ],
    "\u008d": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u008f": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0080": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u008e": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0082": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u008a": [(Encoding.UTF_8, Encoding.ISO_8859_5), (Encoding.CP_866, Encoding.ISO_8859_5)],
    "\u0081": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u008c": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0084": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0091": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0088": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0092": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0090": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u009f": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u009d": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0083": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0097": [(Encoding.UTF_8, Encoding.ISO_8859_5), (Encoding.CP_866, Encoding.ISO_8859_5)],
    "\u0085": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u008b": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0087": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0093": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0086": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0094": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0095": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0098": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0096": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u009e": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u009c": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0099": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0089": [(Encoding.CP_866, Encoding.ISO_8859_5), (Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u00ae": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u0490": [(Encoding.UTF_8, Encoding.WIN_1251), (Encoding.CP_866, Encoding.WIN_1251)],
    "\u00a9": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u00a6": [
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u045f": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.WIN_1251, Encoding.ISO_8859_5),
    ],
    "\u2020": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u2022": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\ufffd": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u2122": [
        (Encoding.CP_866, Encoding.WIN_1251),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u00a8": [
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
    ],
    "\u00a2": [
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.CP_866, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u017d": [(Encoding.CP_866, Encoding.WIN_1252), (Encoding.UTF_8, Encoding.WIN_1252)],
    "\u0160": [(Encoding.UTF_8, Encoding.WIN_1252), (Encoding.CP_866, Encoding.WIN_1252)],
    "\u017e": [(Encoding.CP_866, Encoding.WIN_1252), (Encoding.UTF_8, Encoding.WIN_1252)],
    "\u2260": [(Encoding.UTF_8, Encoding.MAC), (Encoding.CP_866, Encoding.MAC)],
    "\u00b4": [
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.UTF_8, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
    ],
    "\u00b6": [
        (Encoding.UTF_8, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.WIN_1251),
        (Encoding.ISO_8859_5, Encoding.WIN_1252),
        (Encoding.CP_866, Encoding.MAC),
        (Encoding.UTF_8, Encoding.WIN_1251),
        (Encoding.UTF_8, Encoding.WIN_1252),
    ],
    "\u2561": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
    ],
    "\u2562": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u2563": [
        (Encoding.ISO_8859_5, Encoding.CP_437),
        (Encoding.ISO_8859_5, Encoding.CP_866),
        (Encoding.ISO_8859_5, Encoding.KOI8_R),
        (Encoding.UTF_8, Encoding.CP_437),
        (Encoding.UTF_8, Encoding.CP_866),
        (Encoding.UTF_8, Encoding.KOI8_R),
    ],
    "\u0491": [(Encoding.ISO_8859_5, Encoding.WIN_1251), (Encoding.UTF_8, Encoding.WIN_1251)],
    "\u00be": [(Encoding.ISO_8859_5, Encoding.WIN_1252), (Encoding.UTF_8, Encoding.WIN_1252)],
    "\u00b3": [(Encoding.ISO_8859_5, Encoding.WIN_1252), (Encoding.UTF_8, Encoding.WIN_1252)],
    "\u00b9": [(Encoding.ISO_8859_5, Encoding.WIN_1252), (Encoding.UTF_8, Encoding.WIN_1252)],
    "\u222b": [(Encoding.ISO_8859_5, Encoding.MAC), (Encoding.UTF_8, Encoding.MAC)],
    "\u220f": [
        (Encoding.WIN_1251, Encoding.MAC),
        (Encoding.ISO_8859_5, Encoding.MAC),
        (Encoding.UTF_8, Encoding.MAC),
    ],
    "\u2211": [(Encoding.ISO_8859_5, Encoding.MAC), (Encoding.UTF_8, Encoding.MAC)],
    "\u2202": [(Encoding.ISO_8859_5, Encoding.MAC), (Encoding.UTF_8, Encoding.MAC)],
    "\u0459": [(Encoding.UTF_8, Encoding.WIN_1251), (Encoding.WIN_1251, Encoding.ISO_8859_5)],
    "\u02c7": [(Encoding.WIN_1251, Encoding.MAC)],
    "\u02d8": [(Encoding.WIN_1251, Encoding.MAC)],
    "\u009a": [(Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u009b": [(Encoding.UTF_8, Encoding.ISO_8859_5)],
    "\u0161": [(Encoding.UTF_8, Encoding.WIN_1252)],
}
